#
# $Copyright: Copyright (c) 2016 Veritas Technologies LLC.
# All rights reserved.
#
# THIS SOFTWARE CONTAINS CONFIDENTIAL INFORMATION AND TRADE SECRETS OF
# VERITAS TECHNOLOGIES LLC.  USE, DISCLOSURE OR REPRODUCTION IS PROHIBITED
# WITHOUT THE PRIOR EXPRESS WRITTEN PERMISSION OF VERITAS TECHNOLOGIES LLC.
#
# The  Licensed  Software  and  Documentation  are  deemed  to be commercial
# computer  software  as  defined  in  FAR  12.212 and subject to restricted
# rights  as  defined in FAR Section 52.227-19 "Commercial Computer Software
# - Restricted  Rights"  and  DFARS 227.7202,  et seq.  "Commercial Computer
# Software  and  Commercial Computer Software Documentation," as applicable,
# and any successor regulations, whether delivered by Veritas as on premises
# or  hosted  services.  Any  use,  modification,  reproduction  release,
# performance,  display  or  disclosure  of  the  Licensed  Software  and
# Documentation by the U.S. Government shall be solely in accordance with
# the terms of this Agreement. $
#
# $RCSfile: SVCCopyServicesSnap.pm,v $ $Revision: 1.1.2.11 $
#

package SVCCopyServicesSnap;

use ArrayRep;
@ISA = ("ArrayRep");
use POSIX qw(uname);

use SVCCopyServicesAgent;

use strict;
use vars qw($VERSION @ISA @EXPORT @EXPORT_OK %default %vmattrib);
@EXPORT = qw(
);

use ag_i18n_inc;

my $vcs_home=$VCSHOME;
my ($VXDG, $VXDCTL, $VXDISK, $VXDDLADM, $VXASSIST);
my $DEVNULL;

my ($OS) = POSIX::uname();
if ( $^O =~ /^MSWin/i) {
	require Win32;
	Win32->import;

	my $vmpath;
	$vmpath = $ENV{"VMPATH"};
	$vmpath = "\"". Win32::GetShortPathName($vmpath). "\"";
	$VXDG = "$vmpath" . "vxdg";
	$VXDISK = "$vmpath" . "vxdisk";
	$VXASSIST = "$vmpath" . "vxassist";
	
	#$VXDCTL = "$vmpath" . "vxdctl";
	#$VXDDLADM = "$vmpath" . "vxddladm";

	$DEVNULL = "nul";
} else {
	$VXDG = "/usr/sbin/vxdg";
	$VXDCTL = "/usr/sbin/vxdctl";
	$VXDISK = "/usr/sbin/vxdisk";
	$VXDDLADM = "/usr/sbin/vxddladm";
	$DEVNULL = "/dev/null";
}

my $DGTYPE = "DiskGroup";
my $SVCINFO_CMD = "svcinfo";
my $SVCTASK_CMD = "svctask";
my $LSFCCG = "lsfcconsistgrp";
my $LSFCMAP = "lsfcmap";
my $LSVDISK = "lsvdisk";
my $LSFC = "lsfc";
my $PRESTARTFC = "prestartfc";
my $STARTFC = "startfc";
my $STOPFC = "stopfc";
my $STARTRC = "startrc";
my $STOPRC = "stoprc";

my $RECREATEVGTIMER = 15;# If recreatevg fails,  how many times do we retry
my $RECREATEVGINTVL = 5; # If recreatevg fails, interval to wait before retrying

my $snapshot_disks_file;

# Valid states for Flash copy map or consistency group
my @FC_VALID_STATES =  ("preparing",
			"prepared",
			"copying",
                        "idle_or_copied",
			"stopped"
                        );
our $FC_STATE_PREPARED = "prepared";
our $FC_STATE_IDLE_OR_COPIED = "idle_or_copied";
our $FC_STATE_COPYING = "copying";
our $FC_STATE_STOPPED = "stopped";

%default = {
	resname => '',
	
	#
	# For replicated DGs, target is the resname of SVC resource.
	# For non-replicated DGs, target is name of the DiskGroup resource.
	#
	target  => '',

	# Type of snapshot
	sotype  => '',

	# Target is a replicated DG?
	nonrep_dg => '',

	# List of dgs dependent on me.
	dgname  => '',

	# Type of volume manager
	vmtype	=> '',

	# Flash copy map id or name
	fcmapname => '',

	# Is space efficient snapshot
	issesnapshot => '',

	# SVC resource object
	svcres => '',

	#
	# All flash copy cmds have the suffix "map" when invoked for a FC map.
	# When invoked for a FC consistency group, the suffix is "consistgrp"
	# E.g. svcinfo lsfcmap or lsfcconsistgrp
	#
	fc_cmd_suffix => '',

	# Start time - of the entry point
	start_time => '',

	# Possible values: AIX/Solaris/Linux/HPUX
	ostype	=> ''
};

%vmattrib = (
	LVMVG		=> 'VolumeGroup',
	DiskGroup	=> 'DiskGroup',
	CVMVolDg	=> 'CVMDiskGroup',
	VMDg 		=> 'DiskGroupName'	
);

sub new {
	my ($proto, $resname, $target, $mountSO, $useSO, $requireSO, $fcmapname, $issesnapshot) = @_;
	my (@dgname);
	my $svcres_name;
	my @tmp = "";
	my $index = 0;
	my $dgattrib = "";

	my $class = ref($proto) || $proto;
	my $self = { %default };
	$self->{'start_time'} = time();
	$self->{'resname'} = $resname;
	$self->{'target'} = $target;
	if ($OS ne "Windows NT") {
		$self->{'ostype'} = `uname -s`;
	}
	chomp $self->{'ostype'};

	if ($mountSO) {
		if ($requireSO) {
			$self->{'sotype'} = "gold-only";
		} else {
			$self->{'sotype'} = "gold";
		}
	} elsif ($useSO) {
		if ($requireSO) {
			$self->{'sotype'} = "silver-only";
		} else {
			$self->{'sotype'} = "silver";
		}
	} else {
		$self->{'sotype'} = "bronze";
	}


	$self->{'nonrep_dg'} = 0;

	$self->{'fcmapname'} = $fcmapname;

	# Space efficient snapshot is not currently supported in firedrill.
	if ($OS ne "Windows NT") {
		$self->{'issesnapshot'} = $issesnapshot;
		$snapshot_disks_file = sprintf("/var/VRTSvcs/log/tmp/%s.snapshotdisks", $self->{'resname'}); 
	} elsif ($issesnapshot != 0) {
		VCSAG_LOG_MSG( "W", "Space efficient snapshots are not supported on windows. The attribute IsSpaceEfficientSnapshot should be set to 0", 30314);
	}

        # Get all the SVC specific values and VxVM DG name from group variables
        my ($typ) = split ('\n', `$ag_i18n_inc::HARES -value $target Type`);
	if ($typ eq "SVCCopyServices") {

		# Build the list of dependent VxVM DGs
		my (@depfields) = 
			split('\n', `$ag_i18n_inc::HARES -dep $target`);
                foreach my $depline (@depfields) {
                        my (@deppair) = split('\s+', $depline);
                        if ($deppair[2] eq $target) {
                                @tmp = split('\n', 
	    	            `$ag_i18n_inc::HARES -value $deppair[1] Type`);
				if ( 0 == length $self->{'vmtype'} ) {
					$self->{'vmtype'} = $tmp[0];
					$dgattrib = $vmattrib{ $self->{'vmtype'} };
				} else {
					if ( $self->{'vmtype'} ne $tmp[0] ) {
						VCSAG_LOG_MSG( "E", "Mixed volume manager types are not supported; exiting", 30302);
						die("Mixed volume manager types used in configuration");
					}
				}

				@tmp = split('\n', `$ag_i18n_inc::HARES -value $deppair[1] $dgattrib`);
				$dgname[$index] = $tmp[0];
				$index = $index + 1;
				if ($self->{'vmtype'} ne "DiskGroup" && $self->{'vmtype'} ne "LVMVG" && $self->{'vmtype'} ne "VMDg") {
					VCSAG_LOG_MSG("E", "Unsupported type of volume manager $tmp[0]", 30000, "$tmp[0]");
					return undef;
				}
                        }
                }
		$self->{'dgname'} = join(' ', "@dgname");

		# Set SVC resource name
		$svcres_name = $target;

		if (($self->{'vmtype'} eq "LVMVG") && ($self->{'ostype'} !~ /AIX/i)) {
			VCSAG_LOG_MSG("E", "LVM is supported only on AIX; Exiting", 30303);
			return undef;
		}

	} else {
		# Case of non-replicated DiskGroup.
		$self->{'nonrep_dg'} = 1;

		# Get the target VxVM DiskGroup. 
                (@tmp) = split('\n',
                     `$ag_i18n_inc::HARES -value $target DiskGroup`);
		$dgname[$index] = $tmp[0];
		$self->{'dgname'} = join(' ',@dgname);

		# Set SVC resource name (set to any svc res in this VCS group).
		my ($sg) = split('\n', 
				`$ag_i18n_inc::HARES -value $target Group`);
                my (@svcresources) = split('\n',
				`$ag_i18n_inc::HARES -list Type=SVCCopyServices Group=$sg -localclus`);
                (@tmp) = split('\s+', $svcresources[0]);

		$svcres_name = $tmp[0];
	}

	@tmp = split('\n', `$ag_i18n_inc::HARES -value $svcres_name GroupName`);
	my ($groupname) = $tmp[0];

	@tmp = split('\n', `$ag_i18n_inc::HARES -value $svcres_name IsConsistencyGroup`);
	my ($is_consistency_group) = $tmp[0];

	@tmp = split('\n', `$ag_i18n_inc::HARES -value $svcres_name SSHBinary`);
	my ($ssh_bin) = $tmp[0];

	@tmp = split('\n', `$ag_i18n_inc::HARES -value $svcres_name SSHPathToIDFile`);
	my ($ssh_path_to_id_file) = $tmp[0];

	@tmp = split('\n', `$ag_i18n_inc::HARES -value $svcres_name SVCClusterIP`);
	my ($svc_cluster_ip) = $tmp[0];

	@tmp = split('\n', `$ag_i18n_inc::HARES -value $svcres_name SVCUserName`);
	my ($svc_username) = $tmp[0];

	$self->{'svcres'} = SVCCopyServicesAgent->new($svcres_name, $groupname, 
		$is_consistency_group, $ssh_bin, $ssh_path_to_id_file,
		$svc_cluster_ip, $svc_username);
	$self->{'svcres'}->init_values();

	if ($is_consistency_group == 1) {
		$self->{'fc_cmd_suffix'} = "consistgrp";
	} else {
		$self->{'fc_cmd_suffix'} = "map";
	}

	bless($self, $class);
	return $self;
}


sub DESTROY {}

sub _run_svcinfo_cmd {
	my ($self, $sub_cmd, $name) = @_;
	my ($cmd) = "";
	my ($pr_cmd) = "";
	my $ret = 0;
	my @out = ();
	my $ssh_identity_file_option = "";
	my $pr_ssh_identity_file_option = "";
	my $user_name = "admin";
	my $ssh_cmd = "";
	my $svcres = $self->{'svcres'};

	VCSAG_LOGDBG_MSG(2, "Start _run_svcinfo_cmd");

	# Check that the ssh binary is an executable
	if (! -f $svcres->{'ssh_binary'} || ! -x $svcres->{'ssh_binary'}) {
		VCSAG_LOG_MSG("E",
			"The file $svcres->{'ssh_binary'} is not an executable",
			30, "$svcres->{'ssh_binary'}");
		$ret = -1;
		VCSAG_LOGDBG_MSG(2, "End _run_svcinfo_cmd");
		return $ret, @out;
	}

	# If it is specified, check that the id file exists
	if ((! -z $svcres->{'ssh_identity_file'}) && 
		($svcres->{'ssh_identity_file'} ne "")) {
	    if (-e $svcres->{'ssh_identity_file'}) {
		if (-f $svcres->{'ssh_identity_file'}) {
			$ssh_identity_file_option = "-i $svcres->{'ssh_identity_file_w_quotes'}";
			$pr_ssh_identity_file_option = "%-i $svcres->{'ssh_identity_file_w_quotes'}";
		} else {
			VCSAG_LOG_MSG("W",
			"Identity file $svcres->{'ssh_identity_file'} is not an ordinary file",
			31, "$svcres->{'ssh_identity_file'}");
		}
	    } else {
		VCSAG_LOG_MSG("W",
		"Identity file $svcres->{'ssh_identity_file'} does not exist",
			32, "$svcres->{'ssh_identity_file'}");
	    }
	}

	# Check that the user name is specified. (Default is "admin"). 
	if (!-z $svcres->{'svc_username'}) {
		$user_name = $svcres->{'svc_username'};
	}

	$ssh_cmd = "$svcres->{'ssh_binary_w_quotes'}"." "."$ssh_identity_file_option"." -l "."$user_name"." "."$svcres->{'svc_cluster_ip'}";
	my $pr_ssh_cmd = "$svcres->{'ssh_binary_w_quotes'}"." "."$pr_ssh_identity_file_option"." %-l "."$user_name"." "."$svcres->{'svc_cluster_ip'}";

	if ($sub_cmd eq $LSFCCG) {
		# Run the lsfcconsistgrp cmd
		$cmd = $ssh_cmd." ".$SVCINFO_CMD." ".$LSFCCG." -delim : ".$self->{'fcmapname'};
		$pr_cmd = $pr_ssh_cmd." ".$SVCINFO_CMD." ".$LSFCCG." %-delim : ".$self->{'fcmapname'};
	} elsif ($sub_cmd eq $LSFCMAP) {
		# Run the lsfcmap cmd
		$cmd = $ssh_cmd." ".$SVCINFO_CMD." ".$LSFCMAP." -delim : ".$name;
		$pr_cmd = $pr_ssh_cmd." ".$SVCINFO_CMD." ".$LSFCMAP." %-delim : ".$name;
	} elsif ($sub_cmd eq $LSFC) {
		# Run the lsfc cmd
		$cmd = $ssh_cmd." ".$SVCINFO_CMD." ".$LSFC.$self->{'fc_cmd_suffix'}." -delim : ".$self->{'fcmapname'};
		$pr_cmd = $pr_ssh_cmd." ".$SVCINFO_CMD." ".$LSFC.$self->{'fc_cmd_suffix'}." %-delim : ".$self->{'fcmapname'};
	} elsif ($sub_cmd eq $LSVDISK) {
		# Run the lsvdisk cmd
		$cmd = $ssh_cmd." ".$SVCINFO_CMD." ".$LSVDISK." -delim : ".$name;
		$pr_cmd = $pr_ssh_cmd." ".$SVCINFO_CMD." ".$LSVDISK." %-delim : ".$name;
	}

	VCSAG_LOGDBG_MSG(1, "Executing cmd $pr_cmd", "$pr_cmd");

	@out = split('\n', `$cmd`);
	$ret = $?;
	if ($ret != 0) {
		# If it is an SVC error, ret should be 256. In this case, 
		# there should be an error message that begins with the
		# literal CMMVC.
		if ($ret == 256 && scalar @out > 0 && $out[0] !~ /^CMMVC/) {
			# Skip the msg that is not the CMMVC error msg
			shift @out;
		}
		elsif ($ret != 256) {
			# Not an SVC error. Log an appropriate message
			my $error = join(' ', @out);
			VCSAG_LOG_MSG("E",
				"Encountered error ($error) while executing command ($cmd)",
				33, "$error", "$cmd");
		}
	}

	VCSAG_LOGDBG_MSG(2, "End _run_svcinfo_cmd");

	return $ret, @out;
}

sub _run_svctask_cmd {
	my ($self, $sub_cmd, $cmd_args, $cmd_ref) = @_;
	my $cmd = "";
	my ($pr_cmd) = "";
	my $ret = 0;
	my @out = ();
	my $ssh_identity_file_option = "";
	my $pr_ssh_identity_file_option = "";
	my $user_name = "admin";
	my $ssh_cmd = "";
	my $svcres = $self->{'svcres'};

	VCSAG_LOGDBG_MSG(2, "Start _run_svctask_cmd");

	# Check that the ssh binary is an executable
	if (! -f $svcres->{'ssh_binary'} || ! -x $svcres->{'ssh_binary'}) {
		VCSAG_LOG_MSG("E",
			"The file $svcres->{'ssh_binary'} is not an executable",
			30, "$svcres->{'ssh_binary'}");
		$ret = -1;
		VCSAG_LOGDBG_MSG(2, "End _run_svctask_cmd");
		return $ret, @out;
	}

	# If it is specified, check that the id file exists
	if ((! -z $svcres->{'ssh_identity_file'}) && 
		($svcres->{'ssh_identity_file'} ne "")) {
	    if (-e $svcres->{'ssh_identity_file'}) {
		if (-f $svcres->{'ssh_identity_file'}) {
			$ssh_identity_file_option = "-i $svcres->{'ssh_identity_file_w_quotes'}";
			$pr_ssh_identity_file_option = "%-i $svcres->{'ssh_identity_file_w_quotes'}";
		}
		else {
			VCSAG_LOG_MSG("W",
				"Identity file $svcres->{'ssh_identity_file'} is not an ordinary file",
				31, "$svcres->{'ssh_identity_file'}");
		}
	    }
	    else {
		VCSAG_LOG_MSG("W",
			"Identity file $svcres->{'ssh_identity_file'} does not exist",
			32, "$svcres->{'ssh_identity_file'}");
	    }
	}

	# Check that the user name is specified. (Default is "admin"). 
	if (!-z $svcres->{'svc_username'}) {
		$user_name = $svcres->{'svc_username'};
	}

	$ssh_cmd = "$svcres->{'ssh_binary_w_quotes'}"." "."$ssh_identity_file_option"." -l "."$user_name"." "."$svcres->{'svc_cluster_ip'}";
	my $pr_ssh_cmd = "$svcres->{'ssh_binary_w_quotes'}"." "."$pr_ssh_identity_file_option"." %-l "."$user_name"." "."$svcres->{'svc_cluster_ip'}";

	if ($sub_cmd eq $PRESTARTFC) {
		# Run the prestartfcmap/prestartfcconsistgrp cmd
		$cmd = $ssh_cmd." ".$SVCTASK_CMD." ".$PRESTARTFC.$self->{'fc_cmd_suffix'}." ".$cmd_args." ".$self->{'fcmapname'};
		$pr_cmd = $pr_ssh_cmd." ".$SVCTASK_CMD." ".$PRESTARTFC.$self->{'fc_cmd_suffix'}." %".$cmd_args." ".$self->{'fcmapname'};
	} elsif ($sub_cmd eq $STARTFC) {
		# Run the startrc cmd
		$cmd = $ssh_cmd." ".$SVCTASK_CMD." ".$STARTFC.$self->{'fc_cmd_suffix'}." ".$cmd_args." ".$self->{'fcmapname'};
		$pr_cmd = $pr_ssh_cmd." ".$SVCTASK_CMD." ".$STARTFC.$self->{'fc_cmd_suffix'}." %".$cmd_args." ".$self->{'fcmapname'};
	} elsif ($sub_cmd eq $STOPFC) {
		# Run the stopfc cmd
		$cmd = $ssh_cmd." ".$SVCTASK_CMD." ".$STOPFC.$self->{'fc_cmd_suffix'}." ".$cmd_args." ".$self->{'fcmapname'};
		$pr_cmd = $pr_ssh_cmd." ".$SVCTASK_CMD." ".$STOPFC.$self->{'fc_cmd_suffix'}." %".$cmd_args." ".$self->{'fcmapname'};
	}

	VCSAG_LOGDBG_MSG(1, "Executing cmd $pr_cmd", "$pr_cmd");
	$$cmd_ref = $cmd;
	@out = split('\n', `$cmd`);
	$ret = $?;

	VCSAG_LOGDBG_MSG(2, "End _run_svctask_cmd");
	return $ret, @out;
}

#
# Get the current state of flash copy map or consistency group for the
# specified fc map or fc consistency group. The state can be one of the valid
# states defined above or invalid, if we get back something that is not defined
# in the list of valid fc states defined in this module.
#
sub get_fc_state {
	my ($self) = @_;
	my $state = "";

	# Get state
	# svcinfo lsfcmap/lsfcconsistgrp 
	my ($rc, @out) = $self->_run_svcinfo_cmd($LSFC);
	if (!$rc) {
		# Command succeeded
		foreach my $line (@out) {
			if ($line =~ /^status/) {
				my (@stateinfo) = split(':', $line);
				$state = $stateinfo[1];
				last;
			}
		}
	}

	my $state_valid = 0;
	foreach my $stateval (@FC_VALID_STATES) {
		if ($stateval eq $state) {
			$state_valid = 1;
			last;
		}
	}

	if (!$state_valid) {
		$state = "Invalid";
	}

	return $state;
}

#
# This subroutine waits for a period of time for the state of the
# flash copy map to become the state given as argument. The period of time
# that this routine waits is the remaining time left for online ep to timeout
# minus a grace period of one second. The subroutine sleeps 5 seconds, checks
# the status of flash copy map or consistency group. If it is still not in the
# state specified, then stay in the loop.  Else get out of the loop and return
# from the function.
# Returns 0 on success and 1 on failure
#
sub wait_for_fc_state {
	my ($self, $prev_state, $desired_state) = @_;
	my ($oto) = split('\n', `$ag_i18n_inc::HATYPE -value SVCCopyServicesSnap OnlineTimeout`);
	my $cur_time = time();
	my $remaining_time = ($oto - ($cur_time - $self->{'start_time'}));
	my $count = int($remaining_time / 5) - 1;
	if ($count <= 0) {
		return 1;
	}

	my $i = 1; 
	my $state = "";
	while ($i <= $count) {
		$state = $self->get_fc_state();
		if ($state eq $desired_state) {
			VCSAG_LOG_MSG("N",
				"State of flash copy for $self->{'fcmapname'} changed from $prev_state to $state", 30006, "$self->{'fcmapname'}", "$prev_state", "$state");
			last;
		}
		# sleep 5 seconds
		sleep (5);
		$i = $i + 1;
	}

	return 0;
}

sub _run_cmd {
	my ($self, $cmd, $ignore_failure) = @_;
	my $ret = 0;
	my (@out) = ();

	VCSAG_LOGDBG_MSG(1, "Executing cmd $cmd");

	if (1 == $ignore_failure) {
		@out = `$cmd 2>$DEVNULL`;
	} else {
		@out = `$cmd `;
	}
	$ret = $?;

	VCSAG_LOGDBG_MSG(1, "End _run_cmd");

	return $ret, @out;
}

sub take_snapshot {
	my $self = shift;
	my $ret = 0;
	my $ret1 = 0;
	my $i = 0;
	my $state = "";
	my $rc = 0;
	my (@out) = "";
	my $cmd = "";

	VCSAG_LOG_MSG("N", "Taking a flashcopy of $self->{'target'}.", 
      				30001, "$self->{'target'}");

	# Check if the state of FC is already prepared
	$state = $self->get_fc_state();
	if ($state ne $FC_STATE_PREPARED) {

		# If it is space efficient snapshot and the state is copying,
		# the previous online might have not completed. We will stop
		# the flashcopy and then prestart because prestart will fail
		# if the flashcopy is in copying state.
		if (($self->{'issesnapshot'} == 1) && ($state eq $FC_STATE_COPYING)) {
			($ret, @out) = $self->_run_svctask_cmd($STOPFC, "", \$cmd);
			if ($ret != 0) {
				VCSAG_LOG_MSG("E",
					"The flashcopy was in copying state. Attempt to stop flashcopy to enable prestart failed for $self->{'fcmapname'}. Command $cmd failed with error $out[0]", 30315, "$self->{'fcmapname'}", "$cmd", "$out[0]");
				return $ret;
			}
		}
 
		#
		# Sync (prestart) the FC target disks	
		# svctask prestartfcmap/prestartfcconsistgrp name
		#
		VCSAG_LOG_MSG("N", 
			"Issuing PreStart flashcopy for $self->{'fcmapname'}.", 
      			30020, "$self->{'fcmapname'}");

		($rc, @out) = $self->_run_svctask_cmd($PRESTARTFC, "", \$cmd);
		if ($rc != 0) {
			VCSAG_LOG_MSG("E",
			   "Failed to PreStart Flashcopy for $self->{'fcmapname'}. Command $cmd failed with error $out[0]", 
			    30002, "$self->{'fcmapname'}", "$cmd", "$out[0]");
			return $rc;
		} else {

			# Wait till FC maps are in prepared state
			$ret = $self->wait_for_fc_state($state, 
							$FC_STATE_PREPARED);
			$state = $FC_STATE_PREPARED;
			if ($ret) {
				#
				# Timed out waiting for the state of FC 
				# map/consistency group to go to prepared state.
				#
				VCSAG_LOG_MSG("E", "Timed out while waiting for state of flash copy for $self->{'fcmapname'} to go to Prepared state", 30009, "$self->{'fcmapname'}"); 
				return $ret;
			}
		}
	}

	if ($state eq "$FC_STATE_PREPARED") {

		VCSAG_LOG_MSG("N", 
			"Issuing Start flashcopy for $self->{'fcmapname'}.", 
      			30021, "$self->{'fcmapname'}");

		# Create a pitc with startfcmap/startfcconsistgrp name
		($rc, @out) = $self->_run_svctask_cmd($STARTFC, "", \$cmd);
		if ($rc != 0) {
			VCSAG_LOG_MSG("E",
			   "Failed to Start Flashcopy for $self->{'fcmapname'}. Command $cmd failed with error $out[0]", 
			    30003, "$self->{'fcmapname'}", "$cmd", "$out[0]");
			return $rc;
		}
	}

	# If space efficient snapshot, the copy rate is 0% or NOCOPY and so
	# the state will remain in copying. We can go ahead and import and
	# mount the snapshot.
	if ($self->{'issesnapshot'} == 1) {
		$ret = $self->wait_for_fc_state($state, $FC_STATE_COPYING);
		if ($ret) {
			#
			# Timedout waiting for the state of FC map/consistency
			# group to go to copying
			#
			VCSAG_LOG_MSG("E", "Timed out while waiting for state of flash copy for $self->{'fcmapname'} to go to copying state", 30311, "$self->{'fcmapname'}"); 
			return $ret;
		}
	} else {
		#
		# Wait till status of FC map/FC map consistency group is
		# idle_or_copied to ensure that the FC target disks are a
		# complete (100%) copy of the FC source disks.
		#
		$ret = $self->wait_for_fc_state($state, $FC_STATE_IDLE_OR_COPIED);
		if ($ret) {
			#
			# Timedout waiting for the state of FC map/consistency
			# group to go to idle_or_copied
			#
			VCSAG_LOG_MSG("E", "Timed out while waiting for state of flash copy for $self->{'fcmapname'} to go to idle_or_copied state", 30010, "$self->{'fcmapname'}"); 
			return $ret;
		}
	}
	
	VCSAG_LOG_MSG("N", "FlashCopy for $self->{'target'} taken successfully",
			30004, "$self->{'target'}");

	return 0;
}


sub split_replication {
        my ($self, $stop_opt) = @_;
        my $ret = 0;
	my $cmd = "";
	my (@out) = ();
	my $svcres = $self->{'svcres'};

	VCSAG_LOG_MSG("N", "Stopping replication for group $svcres->{'name'} with options $stop_opt.", 30005, "$svcres->{'name'}", "$stop_opt");

	# Given that state is consistent, we can stop replication.
	($ret, @out) = $svcres->_run_svctask_cmd($STOPRC, "$stop_opt", \$cmd);
	if ($ret != 0) {
		my $error = join(' ', @out);
		VCSAG_LOG_MSG("E",
   			"Failed to stop replication for group $svcres->{'name'}. Command $cmd failed with error $error",
		   	30007, "$svcres->{'name'}", "$cmd", "$error");
	} else {
		VCSAG_LOG_MSG("N", "Successfully stopped replication for group $svcres->{'name'}.", 30008, "$svcres->{'name'}");
	}

        return $ret;
}

sub resume_replication {
        my ($self, $start_opt) = @_;
        my $ret = 0;
	my $cmd = "";
	my (@out) = ();
	my $svcres = $self->{'svcres'};

	VCSAG_LOG_MSG("N", "Restarting replication for group $svcres->{'name'} with options $start_opt.", 30011, "$svcres->{'name'}", "$start_opt");

	($ret, @out) = $svcres->_run_svctask_cmd($STARTRC, $start_opt, \$cmd);
	if ($ret != 0) {
		my $error = join(' ', @out);
		VCSAG_LOG_MSG("E",
		  "Failed to restart replication for group $svcres->{'name'}. Command $cmd failed with error $error",
		  30012, "$svcres->{'name'}", "$cmd", "$error");
	} else {
                VCSAG_LOG_MSG("N", "Successfully restarted Replication for group $svcres->{'name'}.", 30013, "$svcres->{'name'}");
	}

        return $ret;
}

sub pitc_gold {
	my ($self) = @_;
	my $ret = 0;

	VCSAG_LOG_MSG("N", 
		      "Initiating GOLD Firedrill on target $self->{'target'}.", 
		      30014, "$self->{'target'}");

	if ($self->{'nonrep_dg'} == 0) {
		$ret = $self->split_replication();
		if ($ret) {
			return 1;
		}
	}

	$ret = $self->take_snapshot();
	if ($ret) {

		# Resume the replication that we just stopped.
		$self->resume_replication("-force");

		if ($self->{'sotype'} eq "gold-only") {
			VCSAG_LOG_MSG("E", "Failed to take a snapshot of the target $self->{'target'}; exiting.", 30015, "$self->{'target'}");
			return 1;
		} else {
			VCSAG_LOG_MSG("W", "Failed to take a snapshot of the target $self->{'target'}; dropping to bronze.", 30016, "$self->{'target'}");
			$ret = $self->pitc_bronze();
			return $ret;
		}
	} 

	my (@fields) = split('\s+', `$ag_i18n_inc::HASYS -nodeid`);
	system("$ag_i18n_inc::HARES -modify $self->{resname} Responsibility $fields[5]");

	# Don't care about the return value.
	$self->resume_replication("-force");

	if ($self->{'vmtype'} eq "LVMVG") {
		$ret = $self->mount_snapshot_lvm("gold");
	} else {
		$ret = $self->mount_diskgroup("gold");
	}

	return $ret;
}

sub pitc_silver {
	my ($self) = @_;
	my $ret = 0;
	my $fdtype = "silver";
	my $svcres = $self->{'svcres'};
	
	VCSAG_LOG_MSG("N", 
		    "Initiating SILVER Firedrill on target $self->{'target'}.",
		    30017, "$self->{'target'}");

	# Split replication with "rw" for the replication target.
	$ret = $self->split_replication("-access");
	if ($ret) {
		return 1;
	}

	$ret = $self->take_snapshot();
	if ($ret) {
		if ($self->{'sotype'} eq "silver-only") {
			#
			# Resume replication specifying the primary.
			#
			if ($svcres->{'role'} eq "master") {
				$ret = $self->resume_replication("-primary aux -force");
			} else {
				$ret = $self->resume_replication("-primary master -force");
			}

			VCSAG_LOG_MSG("E", "Failed to take a snapshot of the target $self->{'target'}; exiting.", 30015, "$self->{'target'}");
			return 1;
		} else {
			#
			# Fall thru to bronze. aka mount the diskgroup.
			#
			$fdtype = "bronze";

			VCSAG_LOG_MSG("W", "Failed to take a snapshot of the target $self->{'target'}; dropping to bronze.", 30016, "$self->{'target'}");
		}
	}

	$ret = $self->mount_diskgroup($fdtype);
	return $ret;
}

sub pitc_bronze {
	my ($self) = shift;
	my $ret = 0;

	VCSAG_LOG_MSG("N", 
		    "Initiating BRONZE Firedrill on target $self->{'target'}.",
		     30018, "$self->{'target'}");

	#
	# Split replication with "rw" for the replication target.
	#
	$ret = $self->split_replication("-access");
	if ($ret) {
		return 1;
	}

	$ret = $self->mount_diskgroup("bronze");
	return $ret;
}

sub pitc {
	my ($self) = @_;
	my $ret = 0;

	if ($self->{'nonrep_dg'} == 0) {
		my $state = $self->{'svcres'}->get_state();
		if ($state ne "consistent_synchronized") {
			VCSAG_LOG_MSG("E", "Firedrill requires replication to be in consistent_synchronized state. Current state is $state. Exiting.",
			30019, "$state");
			return 1;
		}
	}

	if ($self->{'sotype'} =~ /gold/) {
		$ret = $self->pitc_gold();
	} elsif ($self->{'sotype'} =~ /silver/) {
		$ret = $self->pitc_silver();
	} else {
		$ret = $self->pitc_bronze();
	}

	return $ret;
}

#
# For vxvm as the underlying volume manager, the subroutine will clear
# import information on the disks in the device group, update the udid
# of the disks and set an appropriate tag on the disks so that when the
# vxdg import cmd is fired for the target diskgroup and snapshot diskgroup
# with the appropriate cli options, it will succeed.
# Returns: 0 on success
#	   1 on failure
#
sub svccopyservices_prepare_disks {
	my ($self) = shift;
	my ($case) = shift;
	my ($tagname) = shift; # We should be able to return the 
				# tagname value to the caller this way
				# tagname is a pass-by-reference argument
	my $ret = 0;
	my (@snapshot_disks) = ();
	my (@disks) = ();

	my $tag = "SNAP-";
	my $searchdev = "target_vdisk_name";

	# Find flash copy disks that belong to snapshot target
	@snapshot_disks = $self->get_snapshot_disks($searchdev);
	
	# If space efficient snapshot, save the vxvm disks belonging to the
	# snapshot in a file. These have to be removed at the end of offline.
	if ($self->{'issesnapshot'} == 1) {
		open (FILE, "> $snapshot_disks_file");
		foreach my $disk (@snapshot_disks) {
			print FILE $disk."\n";
		}
		close FILE;
	}

	#
	# For gold, we need to prepare the snapshot disks. 
	# For silver and bronze we need to prepare the target disks.
	#
	if ($case ne "gold") {
		# For silver and bronze cases
		$tag = "DEV-";
		$searchdev = "source_vdisk_name";

		# Find flash copy disks that need to be tagged 
		@disks = $self->get_snapshot_disks($searchdev);
	} else {
		@disks = @snapshot_disks;
	}

	# set the tagname passed by reference
	$$tagname = $tag . $self->{'target'};

	my $final_tag = $tag . $self->{'target'};

	# In the future, there should be code to check the type of
	# volume manager technology before calling the appropriate
	# volume-manager-specific routine to set up the disks for an
	# import operation.
	$ret = $self->vxvm_prepare_disks($final_tag, @disks);

	return $ret;
}

sub get_snapshot_disks {
	my ($self) = shift;
	my ($searchdev) = shift;

	my $ret = 0;
	my $rc = 0;
	my ($cnt) = 0;
	my (@fcmap);
	my (@disks) = ();
	my (@out) = ();
	my (@fields) = ();
	my (@vdisk) = ();
	my (@vdisk_uid) = ();
	my ($vd);
	my $line = "";
	my $svcres = $self->{'svcres'};

	# If fcmapconsistgrp, the find the fcmaps.	
	if ($svcres->{'is_consistency_group'} == 1) {

		# svcinfo lsfcconsistgrp -delim : <FC consistency group name>
		$line = "";
		($rc, @out) = $self->_run_svcinfo_cmd($LSFCCG);	
		if (!$rc) {
			# Command succeeded
			foreach $line (@out) {
				# Find FC mapping name
				if ($line =~ /^FC_mapping_name/) {
					@fields = split(':', $line);
					$fcmap[$cnt] = $fields[1];
					$cnt++;
					next;
				}
			}
		}
	} else {
		$fcmap[0] = $self->{'fcmapname'};
	}	
	
	@out = ();
	$cnt = 0;
	my $fc;
	foreach $fc (@fcmap) {
		#
		# Get FC vdisks from fcmap or fcmapconsistgrp
		# svcinfo lsfcmap -delim : <FC map name>
		#
		$line = "";
		($rc, @out) = $self->_run_svcinfo_cmd($LSFCMAP, $fc);	
		
		if (!$rc) {
			# Command succeeded
			foreach $line (@out) {
				# Find vdisk
				if ($line =~ /^${searchdev}/) {
					@fields = split(':', $line);
					$vdisk[$cnt] = $fields[1];
					$cnt++;
					last;;
				}
			}
		}
		next;
	}
		
	# Get uid of vdisk
	my $vdisk_cmd = "";
	$cnt = 0;
	foreach $vd (@vdisk) {
		# Get uid of vdisks
		$line = "";
		@out = ();
		# svcinfo lsvdisk -delim : <virtual disk name>
		($rc, @out) = $self->_run_svcinfo_cmd($LSVDISK, $vd);	
		
		if (!$rc) {
			# Command succeeded
			foreach $line (@out) {
				# Find vdisk uid
				if ($line =~ /^vdisk_UID/) {
					@fields = split(':', $line);
					$vdisk_uid[$cnt] = $fields[1];
					$cnt++;
					last;
				}
			}
		}
		next;
	}

	# Find vxvm disks which match the uids of the vdisks. The VxVM disk's
	# lun serial number field has the uid of the vdisk.
	# First get all the disks and their lun serial number
	my %vmdisk;
	my $uid = "";
	my $disk = "";
	my @disk_info;
	my @lun_info;

	@out = ();
	$line = "";
	if ($OS ne "Windows NT") {
		($rc, @out) = $self->_run_cmd("$VXDISK -p list", 0);
		foreach $line (@out) {
			if ($line =~ /^DISK\s+/) {
				@disk_info = split(':', $line);
				$disk = ltrim($disk_info[1]);
				chomp($disk);
			} elsif ($line =~ /^LUN_SERIAL_NO\s+/) {
				@lun_info = split(':', $line);
				$uid = ltrim($lun_info[1]);
				chomp($uid);
				$vmdisk{"$uid"} = "$disk";
				$disk = "";
				$uid = "";
			}
		}
	} else {
		my @disk_list;
		my @disks;
		my $skip = 1;
		($rc, @out) = $self->_run_cmd("$VXDISK list", 0);
		foreach $line (@out) {
			# Skip the heading line
			if ($skip) {
				$skip = 0;
				next;
			}
			@disk_list = split(' ', $line);
			push(@disks, $disk_list[0]);
		}

		foreach $disk (@disks) {
			($rc, @out) = $self->_run_cmd("$VXDISK diskinfo $disk", 1);
			foreach $line (@out) {
				if ($line =~ /^\s+Serial Number/) {
					@lun_info = split(':', $line);
					$uid = ltrim($lun_info[1]);
					chomp($uid);
					$uid =~ /([0-9a-fA-F]+)/;
					$uid = $1;
					$vmdisk{"$uid"} = "$disk";
					$uid = "";
				}
			}
		}
	}

	# Now get the disks with vdisk's uid matching in the disk array
	$cnt = 0;
	foreach $vd (@vdisk_uid) {
		#
		# Get the VxVM disk which has last 32 characters 
		# of the disk udid matching uid
		#
		$disks[$cnt] = $vmdisk{"$vd"};
		$cnt++;
	}

	return @disks;
}

#
# Arguments to the subroutine:
# ----------------------------
# self: Resource instance
# case: Type of firedrill: gold/silver/bronze
#
# Return Values:
# --------------
# 0 on success
# 1 on failure
#
# Functionality:
# --------------
# The subroutine sets up everything for a successful import of the
# diskgroup on the right set of disks: FlashCopy-luns if gold firedrill and
# Target-luns in case of silver and bronze firedrills.

sub mount_diskgroup {
	my ($self, $case) = @_;
	my $ret = 0;
	my $svcres = $self->{'svcres'};

	# For space efficient snapshots, we need to rescan for the snapshot
	# luns that got created after the snapshot was started.
	if ($self->{'issesnapshot'} == 1) {
		my (@cmds) = ();
		my $ret = 0;

		push(@cmds, "$VXDCTL enable");
		$ret = ar_run_cmds(\@cmds);

		(@cmds) = ();
		push(@cmds, "$VXDDLADM assign names");
		$ret = ar_run_cmds(\@cmds);
	}

	#
	# Clear any import info, update udid, set tag on the appropriate disks.
	# Basically setup the disks for the impending dg import.
	#	
	my $tagname = "";

	if ($case eq "bronze") {
		# For bronze there might not a FC map, so we will call the
		# SVCCopyServices agents' prepare disk which does not
		# require FC map name.
		$ret = $svcres->update_disks_udid();
	} else {
		# tagname will be passed by reference into the prepare_disks sub
		$ret = $self->svccopyservices_prepare_disks($case, \$tagname);
	}

	if ($ret) {
		#
		# An appropriate error message would already have been logged
		# in the sub prepare_disks
		#
		return 1;
	}

	#
	# Test import of the dg.
	# In the future, there should be code to check the type of
	# volume manager technology before calling the appropriate
	# volume-manager-specific routine to test that the import
	# operation succeeds.
	#
	$ret = vxvm_test_diskgroup_import($self->{'dgname'}, $case, $tagname);
	return $ret;
}

#
# Arguments to the subroutine:
# ----------------------------
# dgname: The string comprising of the original diskgroups names 
# (without the _fd appended to the names) separated by a " "
# case: Type of Firedrill: gold/silver/bronze
# tagname: The value of the tag set on the disks on which the diskgroup 
#	is to be imported
#
# Return Values:
# --------------
# 0 on success;
# 1 otherwise and in this case, an appropriate error message is logged
#
# Functionality:
# --------------
# This subroutine checks that after the disks have been setup right,
# the diskgroups can be successfully imported on the system. If the
# subroutine encounters error in attempting to import/deport the
# diskgroups, it returns a failure code (1) along with logging the
# appropriate error message.
#   
sub vxvm_test_diskgroup_import {

	my ($dgnamelist, $case, $tagname) = @_;
	my @dglist = split(' ',$dgnamelist);
	my $newdg;
	my $options = "";
	my (@cmds) = ();
	my $ret = 0;

	if ($case eq "bronze") {
		$options = "-C -o updateid";
	} else {
		$options = "-C -o useclonedev=on -o updateid -o tag=$tagname";
	}

	foreach my $dgname (@dglist) {
		$newdg = "$dgname" . "_fd";

		(@cmds) = ();
		if ($OS ne "Windows NT") {
			push(@cmds, "$VXDG $options -n $newdg import $dgname");
		} else {
			push(@cmds, "$VXDG $options -n $newdg import -g $dgname");
		}

		$ret = ar_run_cmds(\@cmds);
		if ($ret) {
			VCSAG_LOG_MSG("E", "Cannot import disk group $dgname in pitc $case; exiting", 40001, "$dgname", "$case");
			return 1;
		}

		(@cmds) = ();
		# Deport the diskgroup 
		if ($OS ne "Windows NT") {
			push(@cmds, "$VXDG deport $newdg");
			$ret = ar_run_cmds(\@cmds);
			if ($ret) {
				VCSAG_LOG_MSG("E", "Cannot deport disk group $newdg in pitc $case; exiting", 40002, "$newdg", "$case");
				return 1;
			}
		}
	}

	return 0;
}

#
# This subroutine prepares disks assuming VxVM to be the volume manager
# technology. This subrtoutine is called from the array-based
# technology specific prepare_disk subroutine.
# Arguments to the subroutine:
# ----------------------------
# disks: List of disks that the device group has been created upon. To
#	get this list of disks, run the technology specific command to view
#	the device group and the disks underlying it.
# tagname: The tagname to be used to tag the specific volume manager disks.
#	This is the tag that will be used to import the disk group on these
#	disks.
# Return Value:
# -------------
# 0 on success;
# 1 otherwise. An appropriate error message is logged by this
#	subroutine when an error is encountered
# Functionality:
# --------------
# This subroutine does the following tasks:
#	1. Gets the volume manager disks that map to the disks the device
#	group is created on
#	2. For each volume manager disk that comprises the device group, the
#	subroutine does:
#		vxdisk clearimport to clear import information on the volume
#		  manager disks
#		vxdisk updateudid to clear the udid_mismatch flag on the disk
#		vxdisk settag <tagname> to set a tag to uniquely identify the
#		  disks during an import operation
#	If any of these operations fail, the subroutine logs an
#	appropriate error message and returns 1 to indicate a failure.
#
sub vxvm_prepare_disks {
	my ($self, $tagname, @disks) = (@_);
	my (%vxdisk) = ();
	my (@cmds) = ();
	my $ret = 0;

	if ($OS eq "Windows NT") {
		push(@cmds, "$VXASSIST rescan");
		$ret = ar_run_cmds(\@cmds);
	}

	my (@lines) = split('\n', `$VXDISK -eq list`);
	foreach my $l (@lines) {
		my (@fields) = split('\s+', $l);
		$vxdisk{$fields[$#fields]} = $fields[0];
	}
	
	# If space efficient snapshot, vxdctl enable has already been done.
	if ($OS ne "Windows NT" && $self->{'issesnapshot'} == 0) {
		push(@cmds, "$VXDCTL enable");
		$ret = ar_run_cmds(\@cmds);
	}

	foreach my $disk (@disks) {
		(@cmds) = ();
		if ($vxdisk{$disk}) {
			$disk = $vxdisk{$disk};
		}

		if ($OS ne "Windows NT") {
			push(@cmds, "$VXDISK clearimport $disk");
			push(@cmds, "$VXDISK settag $disk $tagname");
		} else {
			push(@cmds, "$VXDISK set tag value=$tagname $disk");
		}
		$ret = ar_run_cmds(\@cmds);
		if ($ret) {
			VCSAG_LOG_MSG("E", 
			"Could not tag disk $disk; exiting;", 40003, "$disk");
			return 1;
		}
	}

	# All has succeeded
	return 0;
}

sub mount_snapshot_lvm {
	my ($self, $case) = @_;
	my $ret = 0;
	my ($vgname, $disk, @disks);
	my @cmds = ();

	# Map VG names to VG ids
	my %vgs_to_id = ();

	my $RECREATEVG = "/usr/sbin/recreatevg";
	my $CHDEV      = "/usr/sbin/chdev";
	my $LQUERYVG      = "/usr/sbin/lqueryvg";

	# First get the SVC Flashcopy disk names for each vg.
	# We ask user to create and import the volume group with flash copy
	# disks. The user then needs to run getpvs action EP which creates a
	# file with the PVIDs of the flash copy disks.
	my @vgnames = split( ' ', $self->{'dgname'} );

	# Construct a hash to map VGname to VG serial id
	ArrayRep::aix_lvm_get_vgs_id(\@vgnames, \%vgs_to_id);

	my @vgids = (values %vgs_to_id);

	foreach $vgname (@vgnames) { 
		my $Lv_Prefix = "fd_";

		my $fdvgname = $vgname . "_fd";

		# Get the PVIDs from the VG PVID file created by getpv action EP
		my $vg_disk_file = sprintf("%s/log/tmp/%s.%s%s.disks", $VCSLOG, $self->{resname}, $vgname, "_fd");
		unless (-f $vg_disk_file) {
			VCSAG_LOG_MSG( "E", "The file with list of flashCopy disks was not found. Import and activate the flashcopy volume group and run the getfcdisks action on the resource.", 30306);
				return 1;
		}

		open (DISKS, $vg_disk_file);
		while ($disk = <DISKS>) {
			chomp($disk);

			my $disk_vgid = `$LQUERYVG -p $disk -v 2`;
			chomp($disk_vgid);

			if ($disk_vgid ne $vgs_to_id{$vgname}) {
				VCSAG_LOG_MSG("E", "The disk $disk in the $vg_disk_file doesn't have the vgid of $vgname", 30307, "$disk", "$vg_disk_file", "$vgname"); 
				if ($vgs_to_id{$vgname} eq "") {
					VCSAG_LOG_MSG("E", "The volume group serial id is not available for $vgname. Ensure that the volume group $vgname is imported.", 30310, "$vgname", "$vgname");
				} 
				return 1;
			}

			push (@disks, $disk);

			push (@cmds, "$CHDEV -l $disk -a pv=clear");
			$ret = ArrayRep::ar_run_cmds(\@cmds);

			if ($ret) {
				VCSAG_LOG_MSG( "W", "Could not clear PV id on $disk. recreatevg might fail", 30308, "$disk" );
			}
		}
		close DISKS;

		# At times, recreatevg fails when called immediately after a
		# split. So, sleep for $RECREATEVGINTVL seconds before trying
		# recreatevg
		for (my $i = $RECREATEVGTIMER ; $i > 0 ; $i--) {
			push (@cmds, "$RECREATEVG -y $fdvgname -Y $Lv_Prefix -n @disks" );
			$ret       = ArrayRep::ar_run_cmds( \@cmds );
			$Lv_Prefix = "NA";
			last if ( !$ret );
			VCSAG_LOG_MSG( "I", "recreatevg failed; retrying again after $RECREATEVGINTVL seconds; $i more retries left", 30309, "$RECREATEVGINTVL", "$i" );
			push(@cmds, "sleep $RECREATEVGINTVL ");
			ArrayRep::ar_run_cmds( \@cmds );
		}

		if ($ret) {
			VCSAG_LOG_MSG( "E", "Cannot recreate VG $fdvgname; exiting", 139, "$fdvgname" );
			return 1;
		}

		my $line = `$ag_i18n_inc::HARES -list VolumeGroup=$fdvgname`;
		my ( $lvmresname, undef ) = split( /\s+/, $line );

		ArrayRep::aix_lvm_update_majornum_attrib($lvmresname, \@disks);
	}

	return 0;
}

sub resynch_gold {
	my ($self) = @_;
	my $ret = 0;
	my $svcres = $self->{'svcres'};

	if ($self->{'nonrep_dg'} == 0) {
		if ($self->{'vmtype'} eq "LVMVG") {
			my @dgs = map ($_ . "_fd", (split ( /\s+/, $self->{'dgname'})));
			$ret = ArrayRep::aix_lvm_exportvgs( \@dgs );
		}

		my ($took) = split('\n', `$ag_i18n_inc::HARES -value $self->{'resname'} Responsibility`);
		if (!$took) {
			#
			# This is gold fall thru to bronze case..
			#
			if ($svcres->{'role'} eq "master") {
				$ret = $self->resume_replication("-primary aux -force");
			} else {
				$ret = $self->resume_replication("-primary master -force");
			}
		}
	}

	return $ret;
}

sub resynch_silver {
	my ($self) = @_;
	my $ret = 0;
	my $svcres = $self->{'svcres'};

	if ($svcres->{'role'} eq "master") {
		$ret = $self->resume_replication("-primary aux -force");
	} else {
		$ret = $self->resume_replication("-primary master -force");
	}

	return $ret;
}

sub resynch_bronze {
	my ($self) = @_;
	my $ret = 0;
	my $svcres = $self->{'svcres'};

	if ($svcres->{'role'} eq "master") {
		$ret = $self->resume_replication("-primary aux -force");
	} else {
		$ret = $self->resume_replication("-primary master -force");
	}

	return $ret;
}

sub resynch {
	my ($self) = @_;
	my $ret = 0;
	my $cmd = "";
	my (@out) = ();
	my $state = "";
	my (@cmds) = ();
	my $ret = 0;

	# If is is a space efficient snapshot, it needs to be stopped, else
	# the next attempt to take snapshot will fail at prestart because
	# state is copying.
	if ($self->{'issesnapshot'} == 1) {
		$state = $self->get_fc_state();
		if ($state ne $FC_STATE_STOPPED) {
			($ret, @out) = $self->_run_svctask_cmd($STOPFC, "", \$cmd);
			if ($ret != 0) {
				VCSAG_LOG_MSG("E",
					"Failed to Stop Flashcopy for $self->{'fcmapname'}. Command $cmd failed with error $out[0]",
					30312, "$self->{'fcmapname'}", "$cmd", "$out[0]");
				return $ret;
			}

			# Confirm that fc map status is stopped.
			$ret = $self->wait_for_fc_state($state, $FC_STATE_STOPPED);
			if ($ret) {
				VCSAG_LOG_MSG("E", "Timed out while waiting for state of flash copy for $self->{'fcmapname'} to go to stopped state", 30313, "$self->{'fcmapname'}");
				return $ret;
			}
		}
	}

	if ($self->{'sotype'} =~ /gold/) {
		$ret = $self->resynch_gold();
	} elsif ($self->{'sotype'} =~ /silver/) {
		$ret = $self->resynch_silver();
	} else {
		$ret = $self->resynch_bronze();
	}

	# If it is space efficient, the snapshot lun goes away after stopping
	# flashcopy. Remove the disks from VxVM because these are now stale
	# entries. Then delete the file that has the list of snapshot disks.
	if ($self->{'issesnapshot'} == 1) {
		my $vmdisk;
		my $keep_snapshot_file = 0;

		open(FILE, $snapshot_disks_file) or die("Could not open file for snapshot disks");
		while (<FILE>) {
			$vmdisk = $_;
			chomp($vmdisk);

			push(@cmds, "$VXDISK rm $vmdisk");
			$ret = ar_run_cmds(\@cmds);	
			if ($ret) {
				VCSAG_LOG_MSG("E",
				"Could not remove the disk $vmdisk from VxVM", 30316, "$vmdisk");
				$keep_snapshot_file = 1;
			}
			$vmdisk = "";
			(@cmds) = ();
		}	 

		close(FILE);

		if (!$keep_snapshot_file) {
			unlink($snapshot_disks_file);
		}

		push(@cmds, "$VXDCTL enable");
		$ret = ar_run_cmds(\@cmds);

		(@cmds) = ();
		push(@cmds, "$VXDDLADM assign names");
		$ret = ar_run_cmds(\@cmds);
	}

	return $ret;
}

sub ar_run_cmds {
    if ($OS ne "Windows NT") {
        return ArrayRep::ar_run_cmds(@_);
    }
	my ($CMDSref) = @_;
	my $ret = 0;
	foreach my $cmd (@{$CMDSref}) {
		$ret = ar_do($cmd);
		if (!defined($ret) || $ret) { last; }
	}
	@{$CMDSref} = ();
	return $ret;
}

sub ar_do {
	my ($cmd, $line, $ret);
	my $exit_code = 0;
	($cmd) = (@_);
	my $cmdText = $cmd;
	$cmdText =~ s/"//g; 
	VCSAG_LOG_MSG("N", "Issuing command: $cmdText", 30317, "$cmdText");
	my $pid = open(INPUT,"$cmd 2>&1 |");
	if (!defined($pid)) {
		#
		# open failed
		#
		my $err = "$!";
		VCSAG_LOG_MSG("E", "Command failed with error $err",
			30318, "$err");
		return undef;
	}

	#
	# open succeeded
	#
	while ($line = <INPUT>) {
		chomp($line);
		$ret = ($ret) ? "$ret\n$line" : $line;
	}
	close INPUT;
	$exit_code = $?;
	if ($exit_code) {
		my $status = $exit_code / 256;
		VCSAG_LOG_MSG("E",
			"Command failed with error: $ret and exit code: $status",
			30320, "$ret", "$status");

	} else {
		VCSAG_LOG_MSG("N", "Command succeeded", 30321);
	}
	return $exit_code;
}

# Left trim function to remove leading whitespace
sub ltrim($)
{
	my $string = shift;
	$string =~ s/^\s+//;
	return $string;
}
